/* 
 * File:   main.c
 * Author: raulhcordeiro
 *
 * Created on April 5, 2011, 12:07 PM
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <libgen.h>
#include <sys/stat.h>
#include <stdint.h>
#include <gst/gst.h>
#include <gst/app/gstappsrc.h>
#include <gst/app/gstappsink.h>
#include <gst/app/gstappbuffer.h>

GMainLoop *loop = NULL;
GstElement *appsrc, *appsink;
int sourceid = 0;
void* videoData = NULL;
size_t videoSize = 0;
uint8_t filebyte[400000];

static int stopapp;
void handler(int sig)
{
    printf("handler(): receives signal %d\n", sig);
    stopapp = 1;
    if (loop != NULL) g_main_loop_quit(loop);
}

static gboolean
bus_call (GstBus     *bus,
          GstMessage *msg,
          gpointer user_data)
{
  switch (GST_MESSAGE_TYPE (msg)) {

    case GST_MESSAGE_EOS:
      g_main_loop_quit(loop);
      break;

    case GST_MESSAGE_ERROR: {
      gchar  *debug;
      GError *error;

      gst_message_parse_error (msg, &error, &debug);
      g_free (debug);

      printf("bus_call(): error[%s]\n", error->message);
      g_error_free (error);

      g_main_loop_quit(loop);
      break;
    }
    default:
      break;
  }

  return TRUE;
}

static gboolean read_data (void* user_data)
{
    GstFlowReturn ret;
    GstBuffer *buffer;
    gboolean ok = TRUE;
    struct timespec ts;
    static int frameid = 0;
    char filepath[200];
    FILE *f;
    int fps = 8;
    struct stat s;

    ts.tv_nsec = (long int)(1000/fps)*1000L*1000L; //período entre frames
    ts.tv_sec = 0;
    nanosleep(&ts, NULL);
    sprintf(filepath, "image.png");
    f = fopen(filepath, "rb");
    if (f == NULL)
    {
      printf("read_data(): fopen[%d]\n", errno);
      ok = FALSE;
      goto end;
    }

    fstat(fileno(f), &s);

    size_t nread = 0;
    do
    {
      size_t nreaded = fread(filebyte+nread, sizeof(uint8_t), s.st_size - nread, f);
      if (nreaded == -1)
      {
        printf("read_data(): fread[errno=%d, file=%s]\n", errno, basename(filepath));
        break;
      }
      else
      {
        nread += nreaded;
      }
    } while (nread < s.st_size && !stopapp);

    fclose(f);

    if (nread < s.st_size)
    {
      printf("read_data(): fopen[%d]\n", errno);
      ok = FALSE;
      goto end;
    }

    buffer = gst_buffer_new_and_alloc(nread);
    if (buffer == NULL)
    {
      printf("read_data(): fail on gst_buffer_new_and_alloc\n");
      ok = FALSE;
      goto end;
    }

    memcpy(GST_BUFFER_DATA (buffer), filebyte, nread);
    GST_BUFFER_SIZE (buffer) = nread;

    g_signal_emit_by_name (appsrc, "push-buffer", buffer, &ret);
    gst_buffer_unref (buffer);

    if (ret != GST_FLOW_OK)
    {
      printf("read_data(): fail on g_signal_emit_by_name [ret=%d]\n", ret);
      ok = FALSE;
      goto end;
    }

    frameid=(frameid+1)%fps;
    if (frameid == 0)
    {
      printf("read_data(): eos!!!\n", ret);
      ok = FALSE;
      goto end;
    }

end:
    if (ok == FALSE)
    {
      g_signal_emit_by_name (appsrc, "end-of-stream", &ret);
    }

    return ok;
}

static void start_feed (GstElement * pipeline, guint size, void* user_data)
{
  if (sourceid == 0)
  {
    printf ("start_feed(): start feeding\n");
    sourceid = g_idle_add ((GSourceFunc) read_data, NULL);
  }
}

static void stop_feed (GstElement * pipeline, void* user_data)
{
  if (sourceid != 0)
  {
    printf ("stop_feed(): stop feeding\n");
    g_source_remove (sourceid);
    sourceid = 0;
  }
}

static void
on_new_buffer_from_source (GstElement * elt, void* user_data)
{
  guint size;
  GstBuffer *buffer;

  if (gst_app_sink_is_eos(GST_APP_SINK (elt)))
  {
    printf ("on_new_buffer_from_source(): eos!!!\n");

    char videoname[10];
    sprintf(videoname, "video.ogg");
    FILE* f = fopen(videoname, "wb");
    if (f == NULL)
    {
      printf("fail on create video file!\n");
    }
    else
    {
      fwrite(videoData, sizeof(uint8_t), videoSize, f);
      fclose(f);
    }

    free(videoData);
    videoData = NULL;
    videoSize = 0;
    return;
  }

  /* get the buffer from appsink */
  buffer = gst_app_sink_pull_buffer (GST_APP_SINK (elt));
  if (buffer == NULL)
  {
    printf ("on_new_buffer_from_source(): fail on gst_app_sink_pull_buffer[errno=%d]\n", errno);
    return;
  }

  /* turn it into an app buffer, it's not really needed, we could simply push
   * the retrieved buffer from appsink into appsrc just fine.  */
  size = GST_BUFFER_SIZE (buffer);
  //g_print ("Pushing a buffer of size %d\n", size);
  void* new = realloc(videoData, videoSize+size);
  if (new == NULL)
  {
    printf ("on_new_buffer_from_source(): fail on realloc[errno=%d]\n", errno);
    goto enrefbuf;
  }

  videoData = new;

  memcpy (videoData + videoSize, GST_BUFFER_DATA (buffer), size);
  videoSize += size;
  
enrefbuf:
  gst_buffer_unref (buffer);
}

/*
 * 
 */
int main(int argc, char** argv)
{
  int fail = 0;
  char pipeline_str[400];
  GstElement *pipeline;
  GstBus *bus;

  signal(SIGINT, handler);
  
  gst_init(NULL, NULL);

  loop = g_main_loop_new (NULL, FALSE);
  if (loop == NULL)
  {
    printf("main(): fail on g_main_loop_new [errno=%d]\n", errno);
    return EXIT_FAILURE;
  }

  sprintf(pipeline_str,
          "appsrc name=appsource caps=\"image/png,framerate=\\(fraction\\)8/1\" ! "
          "gdkpixbufdec ! ffmpegcolorspace ! queue ! videorate ! video/x-raw-yuv,framerate=8/1 ! "
          "theoraenc ! oggmux ! appsink name=appdest");
  
  while(!stopapp)
  {
    pipeline = gst_parse_launch((const gchar *) pipeline_str, NULL);
    if (pipeline == NULL)
    {
      printf("main(): fail on gst_parse_launch [errno=%d]\n", errno);
      break;
    }

    bus = gst_pipeline_get_bus (GST_PIPELINE(pipeline));
    if (bus == NULL)
    {
      printf("main(): fail on gst_pipeline_get_bus [errno=%d]\n", errno);
      fail = 1;
      goto unrefpipe;
    }

    gst_bus_add_watch (bus, bus_call, NULL);
    gst_object_unref (bus);

    appsrc = gst_bin_get_by_name(GST_BIN(pipeline), "appsource");
    if (appsrc == NULL)
    {
      printf("main(): fail on gst_bin_get_by_name [errno=%d]\n", errno);
      fail = 1;
      goto unrefpipe;
    }

    g_signal_connect (appsrc, "need-data", G_CALLBACK(start_feed), NULL);
    g_signal_connect (appsrc, "enough-data", G_CALLBACK(stop_feed), NULL);

    gst_object_unref(GST_OBJECT(appsrc));

    appsink = gst_bin_get_by_name (GST_BIN (pipeline), "appdest");
    g_object_set (G_OBJECT (appsink), "emit-signals", TRUE, "sync", FALSE, NULL);
    g_signal_connect (appsink, "new-buffer",
        G_CALLBACK (on_new_buffer_from_source), NULL);
    g_signal_connect (appsink, "eos",
        G_CALLBACK (on_new_buffer_from_source), NULL);

    gst_object_unref(GST_OBJECT(appsink));

    sourceid = 0;
    GstStateChangeReturn ret = gst_element_set_state (pipeline, GST_STATE_PLAYING);
    if (ret == GST_STATE_CHANGE_FAILURE)
    {
      printf("main(): fail on gst_element_set_state(GST_STATE_PLAYING) [ret=%d, errno=%d]\n", ret, errno);
      fail = 1;
      goto unrefpipe;
    }

    g_main_loop_run(loop);

    ret = gst_element_set_state (pipeline, GST_STATE_NULL);
    if (ret == GST_STATE_CHANGE_FAILURE)
    {
      printf("main(): fail on gst_element_set_state(GST_STATE_NULL) [ret=%d, errno=%d]\n", ret, errno);
      fail = 1;
      goto unrefpipe;
    }

unrefpipe:
    gst_object_unref(GST_OBJECT(pipeline));

    if (fail)
    {
      break;
    }
  }

unrefloop:
  g_main_loop_unref(loop);

  gst_deinit();
  return ( fail ? EXIT_FAILURE : EXIT_SUCCESS);
}

